from .postinit import PostInit

from abc import abstractclassmethod, abstractmethod
from typing import Tuple, Dict, Any, List, Type, Optional
import json

class PropertyMissing(Exception):
    """All required properties must be set"""
    pass


class TooManyProperties(Exception):
    """Too many properties in 'require_any'"""
    pass


def _cls_name(obj: Type[object]) -> str:
    return obj.__class__.__name__

# -----------------------------------------
# -----------------------------------------
class RequireAttrs(PostInit):
    """Require an object to have a list of required attributes after init"""
    @property
    @abstractclassmethod
    def require(self) -> Tuple[str, ...]:
        """This will require you to set attributes a and b
           
            Usage:

                require = ('a', 'b')
        """
        pass

    def __postinit__(self) -> None:
        """Check if every required arg is set"""
        for arg in self.require:
            if arg not in self.__dict__:
                raise PropertyMissing(
                    f"'{_cls_name(self)}' is missing attribute '{arg}'")
            if self.__dict__[arg] is None:
                raise PropertyMissing(
                    f"'{_cls_name(self)}' attribute '{arg}' must be not None")

# -----------------------------------------
# -----------------------------------------
class RequireDictParser(RequireAttrs):
    """Parse dict with required parameters, requires declaration 
    of 'require' and 'require_any' class attributes. Class attribute 
    'infer_numbers' sets if number strings will be convreted to numbers
    """

    infer_numbers = True
    
    @property
    @abstractclassmethod
    def require_any(self) -> Optional[Tuple[Tuple[Tuple[str, ...], int], ...]]:
        """This will require you to set attributes either a or b, 
        and either x & y or x & z or y & z

            Usage:

                require_any = (
                    (("a", "b"), 1),
                    (("x", "y", "z"), 2)
                )
        """
        pass

    @abstractmethod
    def declare_args(self) -> None:
        """ Typing the attributes or setting init values as 'None',
        will raise an error on trying to set a different value
        """
        pass

    @abstractmethod
    def set_args(self) -> None:
        """Set all missing properties from 'require_any' here"""
        pass

    @classmethod
    def from_json(cls, filename: str, encoding: str = "utf-8"):
        with open(filename, "r", encoding=encoding) as json_file:
            data = json.load(json_file)
        return cls(data)

    def __init__(self, import_data: Dict) -> None:
        if not isinstance(import_data, dict):
            raise AttributeError(
                f"'import_data' in {_cls_name(self)} must be a dict")

        self._declaring = True
        self.declare_args()
        self._declaring = False

        if self.require_any:
            for keys, amount in self.require_any:
                self._check_require_set(keys, amount, import_data)
                if len(keys) <= amount:
                    raise TooManyProperties(
                        f"'{_cls_name(self)}': Too many required properties in 'import_data': {amount} in {keys}, max is {len(keys)-1}")

        for key in self.require:
            if key not in import_data:
                raise PropertyMissing(f"'{_cls_name(self)}': Missing property {key} in 'import_data'")

        for key, item in import_data.items():
            if isinstance(item, str) and type(self).infer_numbers:
                try:
                    self.__dict__[key] = int(item)
                except ValueError:
                    try:
                        self.__dict__[key] = float(item)
                    except ValueError:
                        self.__dict__[key] = item
            else:
                self.__dict__[key] = item

        self.set_args()

    def __setattr__(self, name: str, value: Any) -> None:
        try:
            declaring = self._declaring
        except AttributeError:
            declaring = False

        if declaring and (value is not None) and name != "_declaring":
            raise AttributeError(f"'{_cls_name(self)}': can only set values to 'None' in `declare_args` ({name}={value})")
            
        super().__setattr__(name, value)

    def __postinit__(self):
        """Check if every required arg is set"""
        super().__postinit__()

        for arg in self._required_any_keys:
            if arg not in self.__dict__:
                raise PropertyMissing(
                    f"'{_cls_name(self)}' is missing attribute '{arg}'")

            if self.__dict__[arg] is None:
                raise PropertyMissing(
                    f"'{_cls_name(self)}' attribute '{arg}' must be not None")

    @property
    def _required_any_keys(self):
        """Returns all property keys in require_any"""
        if self.require_any:
            return [y for x in self.require_any for y in x[0]]
        else:
            return []

    def _check_require_set(self, keys: List, amount: int, import_data: Dict):
        """Checks if data contains correct combination of keys from require_any"""
        if amount < 1:
            raise AttributeError(f"'{_cls_name(self)}': amount of properties for {keys} must be greater or equal to 1")

        found = 0
        for key in keys:
            if key in import_data:
                found += 1

        if not found:
            raise PropertyMissing(f"'{_cls_name(self)}': Missing one of the properties from {keys} in 'import_data'")

        if found > amount:
            raise TooManyProperties(
                f"'{_cls_name(self)}': The required amount of properties in {keys} is {amount} but {found} was found in 'import_data'")



# -----------------------------------------
# ------------Require decorator factory----
# -----------------------------------------
from functools import wraps
from typing import Callable, Type, Tuple, Dict, Any, Optional


def _mangle_args(self: Type[object], inp: str) -> str:
    """Mangle names starting with 'self.__'

    Args:
        self: subclass of object
        inp: str, contains valuate expression
    """
    return inp.replace("self.__", f"self._{_cls_name(self)}__")

def _mangle_single(self: Type[object], inp: str) -> str:
    """Mangle names starting with '__'
    
    Args:
        self: subclass of object
        inp: str, contains name of an object attribute
    """
    return inp.replace("__", f"_{_cls_name(self)}__") if inp.find("__") == 0 else inp

def _cls_name(obj: Type[object]) -> str:
    return obj.__class__.__name__
# -----------------------------------------
# -----------------------------------------

def require_condition(var: Optional[str] = None, cond: Optional[str] = None, msg: Optional[str] = None) -> Callable:
    """Decorator factory for requiring a condition to be True, if 'cond' is None evaluate 'var' alone
    
    Args:
        var: str, argument name is a class
        cond: str, evaluate condition used by eval(), left side is a value of 'var'
        msg: str, display on error
    """
    if var is None and cond is None:
        raise ValueError("at least one argument must be provided: 'var', 'cond'")

    for x in (var, cond, msg):
        if x is not None and not isinstance(x, str):
            raise TypeError(f"argument '{x}' must be str")


    cond_str = cond if cond is not None else ""
    var_str = f"self.{var}" if var is not None else ""
    msg = msg if msg else f"'{var_str}{cond_str}' must evaluate to 'True'"

    def require_decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self: Type[object], *args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
            nonlocal cond, var
            cond = _mangle_args(self, cond) if cond is not None else None
            var = _mangle_single(self, var) if var is not None else None

            try:
                value = self.__dict__[var] if var is not None else ""
            except KeyError:
                raise AttributeError(f"'{_cls_name(self)}' object has no attribute '{var}'") 

            condition = eval(f"{value}{cond}") if cond is not None else bool(value)
            if not condition:
                raise RuntimeError(msg)
                
            return func(self, *args, **kwargs)

        return wrapper

    return require_decorator

# -----------------------------------------
# -----------------------------------------

def once(_func=None, *, mode="cashed"):
    """Decorator to allow a function to only be called once or
     always return result of the 1st call
    
    Args:
        mode: str, 'cashed' - always return 1st result, 
        'error' - raise an error if called a 2nd time
    """
    if mode not in ("cashed", "error"):
        raise ValueError("'mode' argument can only be: 'cashed' or 'error'")
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not wrapper._done:
                wrapper._cashed_result = func(*args, **kwargs)
                wrapper._done = True
            elif mode == "error":
                raise RuntimeError(f"'{func.__name__}' can only be called once")
            return wrapper._cashed_result

        wrapper._cashed_result = None   
        wrapper._done = False
        return wrapper

    if _func is None:
        return decorator
    else:
        return decorator(_func)

# -----------------------------------------
# -----------------------------------------

def limit(max_calls):
    """Decorator to allow a function to only be called 
    'max_calls' amount of times
    
    Args:
        max_calls: int, how many time a function can be called 
    """
    if not isinstance(max_calls, int):
        raise TypeError("'max_calls' must be int")
    if max_calls < 1:
        raise ValueError("'max_calls' must be greater than 0")

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            wrapper._calls += 1
            if wrapper._calls > max_calls:
                raise RuntimeError(f"'{func.__name__}' can only be called {max_calls} times")
            return func(*args, **kwargs)

        wrapper._calls = 0   
        return wrapper

    return decorator

def try_times(times: int):
    """Tries to call function several times if it raises an error
    
    Args:
        times: int, how many times to call
    """
    if not isinstance(times, int):
        raise TypeError("'times' must be int")
    if times < 1:
        raise ValueError("'times' must be greater than 0")

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            res = None
            nonlocal times
            calls = 0
            while calls < times:
                calls += 1
                try:
                    res = func(*args, **kwargs)
                except Exception as err:
                    if calls == times:
                        orig_msg = str(err)
                        raise type(err)(f"function '{func.__name__}' was called {times} times but raised exception: {orig_msg}") from None
                else:
                    break
                
            return res
        return wrapper
    return decorator

if __name__ == "__main__":
    pass


# -----------------------------------------
# ------------Postinit---------------------
# -----------------------------------------
from typing import Tuple, Dict, Any
from abc import abstractmethod, ABCMeta

class PostInitMeta(ABCMeta):
    """Metaclass that adds a call to __postinit__ which is called after __init__"""
    def __call__(cls, *args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
        obj = type.__call__(cls, *args, **kwargs)
        obj.__postinit__()
        return obj
    

class PostInit(metaclass=PostInitMeta):
    """Base class that adds a call to __postinit__ which is called after __init__"""
    @abstractmethod
    def __postinit__(self) -> None:
        pass

# ==============================================


if __name__ == "__main__":
    pass



from queue import Queue
from threading import Thread
from typing import Callable, Any, Type, Optional

def _cls_name(obj: Type[object]) -> str:
    return obj.__class__.__name__

class MessageThread:
    """Runs a thread with queue of messages"""
    def print(self, *args, **kwargs):
        """Print function used for printing out, is standard python print() by default"""
        if not self.__msg_handler.is_alive():
            raise RuntimeError(f"'{_cls_name(self)}' messaging thread is not alive")
        self.__msg_queue.put((args, kwargs)) 

    def __init__(self, print_func: Optional[Callable[..., Any]] = None) -> None:
        self.__msg_queue = Queue()
        self.__print = print_func if print_func is not None else print
        self.__msg_handler = Thread(target=self.__msg_ouput, daemon=True)
        self.__msg_handler.start()
        

    def __msg_ouput(self):
        while True:
            args, kwargs = self.__msg_queue.get()
            self.__print(*args, **kwargs)
            self.__msg_queue.task_done()

    def join(self):
        """Wait for all messages to print out"""
        self.__msg_queue.join()




from typing import List, Tuple, Union
from collections import defaultdict
import re
from pprint import pprint

class Splicer:
    """Specify blocks of text in files with special tags and paste them
    in other files with a different tag
    
        Usage:

            # make block:
            splice@:block1
            some text
            /splice@:block1
            
            # insert block here
            insert@:block1
    """

    def __init__(self, parse_prefix="splice@:", splice_prefix="insert@:") -> None:
        """Args:

            parse_prefix: str, prefix for starting a new block
            splice_prefix: str, prefix for pasting a block
        """
        
        self._parse_prefix = parse_prefix
        self._end_parse_prefix = f"/{self._parse_prefix}"
        self._splice_prefix = splice_prefix 
        self._segments = {}

    def __str__(self) -> str:
        return self._segments

    def parse_batch(self, files: List[str], encoding="utf-8"):
        """Parse a bunch of files at once
        
        Args:
            files: list, a list of files
            encoding: str, file encoding
        """
        for f in files:
            self.parse_segments(f, encoding=encoding)

    def splice_batch(self, files: List[Tuple[str, str]], encoding="utf-8"):
        """Splice a bunch of files at once
        
        Args:
            files: list, a list of tuples with 'filename' and 'out_filename' args 
            encoding: str, file encoding
        """
        for (f_in, f_out) in files:
            self.splice_into(f_in, f_out, encoding=encoding)

    def parse_segments(self, filename, encoding="utf-8"):
        """Parse a file for blocks

        Args:
            filename: str, path to a file
            encoding: str, file encoding
        """
        
        new_segments = defaultdict(list)
        current_segments = set()
        with open(filename, "r", encoding=encoding) as f:
            for line in f:
                if self._parse_prefix in line and self._end_parse_prefix not in line:  
                    re_seg = rf"(?<={self._parse_prefix})(.*)"
                    new_seg = re.findall(re_seg, line)[0]  
                    if new_seg in current_segments:
                        raise RuntimeError(f"'{type(self).__name__}': key '{new_seg}' appeared twice")
                    if new_seg in self._segments:
                        raise RuntimeError(f"'{type(self).__name__}': key '{new_seg}' is already present")
                    current_segments.add(new_seg)
                    continue
                if self._end_parse_prefix in line:
                    re_seg = rf"(?<={self._end_parse_prefix})(.*)"
                    end_seg = re.findall(re_seg, line)[0]
                    if end_seg not in current_segments:
                        raise RuntimeError(f"'{type(self).__name__}': cannot close never started segment '{end_seg}'")
                    current_segments.remove(end_seg)
                    continue

                for key in current_segments:
                    new_segments[key].append(line)
            if len(current_segments) != 0:
                raise RuntimeError(f"'{type(self).__name__}': segments were never closed {current_segments}")

        self._segments = self._segments | new_segments
    def add_segment_manual(self, key: str, text: Union[str, List[str]], overwrite=False):
        """Manually add a segment
        
        Args:
            key: str, segment name
            text: str or List[str], segment content
            overwrite: bool, allow overwrites of existing keys
        """
        if not isinstance(key, str):
            raise TypeError(f"'{type(self).__name__}': 'key' must be str")
        if not isinstance(key, (str, list)):
            raise TypeError(f"'{type(self).__name__}': 'text' must be str or a list")
        if key in self._segments and not overwrite:
            raise RuntimeError(f"'{type(self).__name__}': key '{key}' is already present")
        
        text = text if isinstance(text, list) else text.split("\n")
        if not text[-1].endswith("\n"):
            text[-1] = text[-1] + "\n"

        self._segments[key] = text

    def splice_into(self, filename: str, out_filename: str, encoding="utf-8"):
        """Splice parsed blocks into a file
        
        Args:
            filename: str, path to a file
            out_filename: str, path to an output file
            encoding: str, file encoding
        """
        result_text = []
        with open(filename, "r", encoding=encoding) as f:
            for line in f:
                if self._splice_prefix in line:
                    re_seg = rf"(?<={self._splice_prefix})(.*)"
                    seg_name = re.findall(re_seg, line)[0]
                    if seg_name not in self._segments:
                        raise KeyError(f"'{type(self).__name__}': no '{seg_name}' found")
                    result_text += self._segments[seg_name]
                    continue
                result_text.append(line)
                
        with open(out_filename, "w", encoding=encoding) as f:
            f.writelines(result_text)

    def clear_segments(self):
        """Delete all parsed segments"""
        self._segments = {}



# -----------------------------------------
# ------------Immutable properties---------
# -----------------------------------------
from typing import Any, Type
class PropertyImmutable(Exception):
    """Cannot mutate an already set property"""
    pass

def _cls_name(obj: Type[object]) -> str:
    return obj.__class__.__name__
    
class ImmutableProperties():
    """Immutable properties in class except with prefixes defined in 'mut_prefix'"""

    mut_prefix = ("__", )

    def __setattr__(self, name: str, value: Any) -> None:
        """After the value is set to not None it cannot be changed"""
        cls = type(self)
        mangle = lambda x: f"_{_cls_name(self)}{x}" if x.startswith("__") else x

        if cls.mut_prefix:
            is_mut = any([name.startswith(mangle(t)) for t in cls.mut_prefix])
        else:
            is_mut = False

        if not is_mut:
            try:
                current_value = self.__dict__[name]
            except:
                current_value = None
            if name in self.__dict__ and current_value is not None:
                raise PropertyImmutable(f"Property '{name}' in '{_cls_name(self)}' is immutable")
        self.__dict__[name] = value

if __name__ == "__main__":
    pass


from queue import Queue
from threading import Thread, Event, Lock
from functools import wraps
from typing import List

class _RepeaterResult:

    def __init__(self, func, timer, get_nones, get_false, *args, **kwargs):
        self._func = func
        self._timer = timer
        self._get_nones = get_nones
        self._get_false = get_false
        self._queue = Queue()
        self._thread = Thread(target=self.__run, daemon=True, args=args, kwargs=kwargs)
        self._wait = Event().wait
        self._lock = Lock()
        self._running = True
        self._thread.start()

    def __run(self, *args, **kwargs):
        while self._running:
            result = self._func(*args, **kwargs)
            put_nones = bool((result is not None) or self._get_nones)
            put_non_nones = bool(result or self._get_false)
            if put_nones and put_non_nones:
                self._queue.put(result)
            self._wait(self._timer)

    def get(self) -> List:
        """Get accumulated results from queue"""
        res = []
        self._lock.acquire()
        while not self._queue.empty():
            res.append(self._queue.get())
            self._queue.task_done()
        self._lock.release()
        return res

    def stop(self):
        """Stop timer"""
        self._running = False
        self._thread.join()


def repeat_timer(timer, get_nones=False, get_false=True):
    """Launches a thread that will execute a function each 'timer' sec, 
    to get accumulated results call 'get' method, to stop timer call 'stop' method 
    of returned value

    Args:
        timer: float, delay between calls
        get_nones: bool, put 'None' values in result
        get_false: bool, put bool() == False values in result
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return _RepeaterResult(func, timer, get_nones, get_false, *args, **kwargs)
    
        return wrapper
    return decorator


# from .immutable import ImmutableProperties
# from .postinit import PostInit
# from .require import RequireAttrs, RequireDictParser
# from .message import MessageThread

# __all__ = [
#     ImmutableProperties, 
#     PostInit, 
#     RequireAttrs, 
#     RequireDictParser,
#     MessageThread
# ]



from functools import wraps
from typing import Callable, Tuple, Dict, Any
import time


def time_exec(_func=None, *, printout: bool = True) -> Callable:
    """Decorator factory, adds 'last_call_elapsed' and 'total_time' attributes to a fucntion

    Args:
        printout: bool, printout to console; default is True
    """
    def timing_base(fun: Callable) -> Callable:
        """decorator for timing"""
        @wraps(fun)
        def wrapper(*args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
            start = time.time()
            result = fun(*args, **kwargs)
            end = time.time()
            elapsed = (end-start)*1000.0
            wrapper.last_call_elapsed = elapsed
            wrapper.total_time = wrapper.total_time + elapsed
            if printout:
                print(f'{fun.__name__:s} function took {elapsed:.3f} ms')
            return result

        wrapper.last_call_elapsed = 0
        wrapper.total_time = 0
        return wrapper
        

    if _func is None:
        return timing_base
    else:
        return timing_base(_func)




def postinit(cls_orig):
    """Adds a call to __postinit__ after init was completed"""
    orig_init = cls_orig.__init__

    def __init__(self, *args, **kwargs):
        
        orig_init(self, *args, **kwargs) 
        self.__postinit__()

    cls_orig.__init__ = __init__ 

    return cls_orig